/*
    ball-dE2.cpp - Mlut and DT computations for Euclidean distance in 2D

    Original paper and sources:
      - E. Remy and E. Thiel. Exact Medial Axis with Euclidean Distance. 
        Image and Vision Computing, 23(2):167-175, 2005. 
        https://pageperso.lis-lab.fr/edouard.thiel/IVC2004/index.html

    CC BY-SA Edouard.Thiel@univ-amu.fr, Eric.remy@univ-amu.fr - 04/08/2023
*/

#include "ball-dE2.hpp"


/*
 * Fast Cone Distance Transform Algorithm.
 *
 * Input:  L the side length.
 * Output: CTg the L^2 distance image to the origin.
*/

void Ball_dE2::compute_CTg (int L, int *CTg)
{
    int x, y;

    CTg[0] = 0;

    for (x = 1; x < L; x++)
    for (y = 0; y <= x; y++)
    {
        CTg[ y*L + x ] = x*x + y*y;
    }
}


/*
 * Fast Distance Transform in G(Z^2).
 *
 * Input:  L the side length,
 *         CTg the distance cone to the origin,
 *         R the radius of the ball in CTg.
 * Output: DTg the distance image to the background of the ball.
 *
 * Algorithm derived from Hirata's SEDT, with improvements (mark unpropagated 
 * distances with -1 in cones, stop loops ASAP., store intersections).
 * See: T. Hirata, "A unified linear-time algorithm fo computing distance maps",
 * in Information Processing Letters 58:129-133, 1996.
*/


double D_Intersec (int u, int gu, int v, int gv)
{
    return (u+v + ((double)gu-gv)/(u-v)) / 2;
}

void Ball_dE2::compute_DTg (int L, int *CTg, int *DTg, int R)
{
    int x, xM, y, p, dp, k, propag;
    int Si[L], Sv[L], Sn;  /* index, values, number */
    double Sr[L];          /* intersections */
 
    /* compute bound xM[ and verify that xM < L */
    for (xM = 0; xM < L; xM++) if (CTg[0*L+xM] > R) break;
    if (xM >= L) printf ("WARNING xM is not < L\n");
    
    /* First scan: x++, y-- */
    for (x = 0; x <= xM; x++)
    {
        k = 0; propag = 0;
        for (y = x; y >= 0; y--)
        {
            p = y*L+x;
            if (CTg[p] > R)   /* outside the ball : background */
                propag = 1;
            else if (propag)  /* inside the ball, mark to dist. k*k from bg */
              { k++; DTg[p] = k*k; }
            else DTg[p] = -1; /* inside the ball, no distance propagated */
        }
    }
    
    /* Second and final scan: y++, x++ */
    for (y = 0; y <= xM; y++)
    {
        Sn = 0;  /* empty stacks */
        
        /* Compute stacks indices Si[Sn]=x and values Sv[Sn]=DTg[x,y] */
        for (x = y; x <= xM; x++)
        {
            p = y*L+x; dp = DTg[p];
            if (dp < 0) continue;  /* Non propagated value */
            
            /* To speedup algorithm, stop at the second consecutive 0 */
            if (dp == 0 && x > y && DTg[p-1] == 0) break;
            
            while (Sn >= 2 && D_Intersec (Si[Sn-1], Sv[Sn-1], x, dp) < Sr[Sn-1])
                Sn--;  /* pop */
                  
            /* push */    
            Si[Sn] = x; Sv[Sn] = dp; 
            if (Sn >= 1) Sr[Sn] = D_Intersec (Si[Sn-1], Sv[Sn-1], Si[Sn], Sv[Sn]);
            Sn++;
        }
        
        if (Sn == 0) continue;  /* Empty stack */
        
        /* Compute new DTg values using stacks */
        for (x = xM; x >= y; x--)
        {
            p = y*L+x;
            if (DTg[p] == 0) continue;
            
            while (Sn >= 2 && x < Sr[Sn-1]) Sn--;  /* pop */
                  
            DTg[p] = (x-Si[Sn-1])*(x-Si[Sn-1]) + Sv[Sn-1];
        } 
    }
}


/*
 * Compute the greatest verifiable radius of balls.
 *
 * Input:  L the side length.
 * Output: returns the greatest verifiable radius in the image.
*/

int Ball_dE2::get_greatest_radius (int L)
{
    int res = (L-1)*(L-1)-1;

    printf ("GreatestRadius for L = %d is %d\n", L, res);
    
    return res;
}


//----------------------------- PART USING OPENCV -----------------------------


int Ball_dE2::min_column_SEDT (int col[], int y, int rows)
{
    int min = col[y], k2, d2;

    for (int k = 1; y-k >= 0; k++) {
        k2 = k*k; if (k2 >= min) break;
        d2 = col[y-k] + k2; if (d2 < min) min = d2;
    }
    d2 = 0 + (y+1)*(y+1);           // exterior is considered to be 0
    if (d2 < min) min = d2;

    for (int k = 1; y+k < rows; k++) {
        k2 = k*k; if (k2 >= min) break;
        d2 = col[y+k] + k2; if (d2 < min) min = d2;
    }
    d2 = 0 + (rows-y)*(rows-y);     // exterior is considered to be 0
    if (d2 < min) min = d2;

    return min;
}


void Ball_dE2::compute_SEDT_Saito_Toriwaki (cv::Mat &img_int)
{
    CHECK_MAT_TYPE(img_int, CV_32SC1);

    // Scan of the lines
    for (int y = 0; y < img_int.rows; y++)
    {
        // Forward
        int k = 0;      // exterior is considered to be 0
        for (int x = 0; x < img_int.cols; x++)
        {
            if (img_int.at<int>(y,x) == 0) 
                k = 0;
            else { 
                k++; 
                img_int.at<int>(y,x) = k*k; 
            }
        }

        // Backward
        k = 0;      // exterior is considered to be 0
        for (int x = img_int.cols-1 ; x >= 0; x--)
        {
            if (img_int.at<int>(y,x) == 0) 
            k = 0;
            else {
                k++; 
                int k2 = k*k;
                if (img_int.at<int>(y,x) > k2) img_int.at<int>(y,x) = k2; 
            }
        }
    }

    // Scan of the columns
    for (int x = 0; x < img_int.cols; x++)
    {
        int col[img_int.rows];
        for (int y = 0; y < img_int.rows; y++)
            col[y] = img_int.at<int>(y,x);

        for (int y = 0; y < img_int.rows; y++)
            img_int.at<int>(y,x) = min_column_SEDT (col, y, img_int.rows);
    }
}


int Ball_dE2::max_buffer_Saito_rev (int buf[], int pos, int cmax)
{
    int k, h, max = buf[pos];
    
    for (k = 1; pos-k >= 0; k++) {
        h = buf[pos-k]; if (h == 0) continue;
        h -= k*k; if (h > max) max = h;
    }

    for (k = 1; pos+k < cmax; k++) {
        h = buf[pos+k]; if (h == 0) continue;
        h -= k*k; if (h > max) max = h;
    }

    return max;
}


void Ball_dE2::compute_RSEDT_Saito_Toriwaki (cv::Mat &img_int)
{
    CHECK_MAT_TYPE(img_int, CV_32SC1);

    int buf[std::max(img_int.cols, img_int.rows)];

    // Scan of the lines
    for (int y = 0; y < img_int.rows; y++)
    {
        // Copy line into buf
        for (int x = 0; x < img_int.cols; x++)
            buf[x] = img_int.at<int>(y,x);

        // Calculate max in buf and store in image
        for (int x = 0; x < img_int.cols; x++)
            img_int.at<int>(y,x) = max_buffer_Saito_rev (buf, x, img_int.cols);
    }

    // Scan of the columns
    for (int x = 0; x < img_int.cols; x++)
    {
        // Copy column into buf
        for (int y = 0; y < img_int.rows; y++)
            buf[y] = img_int.at<int>(y,x);

        // Calculate max in buf and store in image
        for (int y = 0; y < img_int.rows; y++)
            img_int.at<int>(y,x) = max_buffer_Saito_rev (buf, y, img_int.rows);
    }
}


void Ball_dE2::compute_level_curves (cv::Mat &img_int)
{
    for (int y = 0; y < img_int.rows; y++)
    for (int x = 0; x < img_int.cols; x++)
        if (img_int.at<int>(y,x) > 0) 
            img_int.at<int>(y,x) = sqrt(img_int.at<int>(y,x));
}



